from setux.logger import debug
from setux.core.manage import SpecChecker


class Distro(SpecChecker):
    manager = 'group'

    def get(self):
        group = self.key if self.key else self.spec['gid']
        ret, out, err = self.run(f'grep ^{group}: /etc/group')
        for line in out:
            name, x, gid, users = line.split(':')
            if self.key:
                if name != self.key: continue
            else:
                if int(gid) != self.spec['gid']: continue
            return dict(
                name = name,
                gid = int(gid),
                users = users.split(','),
            )

    @property
    def name(self):
        return self.get()['name']

    @property
    def gid(self):
        return self.get()['gid']

    def cre(self, check=True):
        debug(f'group create {self.key}')
        self.do_cre(group=self.key, check=check)

    def mod(self, key, val, check=True):
        debug(f'group {self.key} change {key} -> {val}')
        self.do_mod(key=key, val=val, group=self.key, check=check)

    def rm(self, check=True):
        debug(f'group delete {self.key}')
        self.do_rm(group=self.key, check=check)

    def do_cre(self, *, group, check):
        self.run(f'groupadd {group}', check=check)

    def do_mod(self, *, group, key, val, check):
        self.run(f'groupmod --{key} {val} {group}', check=check)

    def do_rm(self, *, group, check):
        self.run(f'groupdel {group}', check=check)


class FreeBSD(Distro):
    opts = dict(
        gid = 'g',
    )

    def do_cre(self, *, group, check):
        self.run(f'pw groupadd -n {group}', check=check)

    def do_mod(self, *, group, key, val, check):
        opt = self.opts[key]
        self.run(f'pw groupmod -n {group} -{opt} {val}', check=check)

    def do_rm(self, *, group, check):
        self.run(f'pw groupdel -n {group}', check=check)




from pybrary.ascii import oct_mod

from setux.logger import debug, error
from setux.core.manage import SpecChecker


class Distro(SpecChecker):
    manager = 'file'

    def get(self):
        ret, out, err = self.run(f'ls -l --color=never {self.key}')
        try:
            mod, ln, usr, grp, size, month, day, time, path = out[0].split()
            typ, mod = mod[0], mod[1:10]
            assert typ=='-', f'FILE {self.key} : {typ} !'
            return dict(
                name = path,
                mod = oct_mod(mod),
                usr = usr,
                grp = grp,
                size = int(size),
            )
        except Exception as x:
            error('file %s !\n%s : %s', self.key, type(x), x)

    def __str__(self):
        data = self.get()
        if data:
            return f'File {self.key} {data["size"]} {data["mod"]} {data["usr"]}:{data["grp"]}'
        else:
            return f'File {self.key} not found'

    def cre(self):
        debug(f'file create {self.key}')
        self.run(f'touch {self.key}')

    def mod(self, key, val):
        debug(f'file {self.key} change {key} -> {val}')
        if key=='mod':
            self.run(f'chmod {val} {self.key}')
        elif key=='usr':
            self.run(f'chown {val} {self.key}')
        elif key=='grp':
            usr = self.spec["usr"]
            self.run(f'chown {usr}:{val} {self.key}')

    def rm(self):
        debug(f'file delete {self.key}')
        self.run(f'rm {self.key}')

    @property
    def size(self):
        return self.get().get('size')

    def read(self):
        content = self.target.read(self.key)
        assert self.size==len(content), f'\n size ! {self.size} ! {len(content)} !'
        return content

    def write(self, content):
        self.target.write(self.key, content)
        assert self.size==len(content), f'\n size ! {self.size} ! {len(content)} !'

    @property
    def hash(self):
        if self.get():
            debug(f'file hash {self.key}')
            ret, out, err = self.run(f'md5sum {self.key}')
            return out[0].split()[0]




from setux.core.package import CommonPackager


# pylint: disable=no-member


class Distro(CommonPackager):
    manager = 'npm'

    def ls(self):
        ret, out, err = self.run('npm list -g --depth=0')
        for line in self.out[1:]:
            if len(line)<5: continue
            n, v = line[4:].split('@')
            yield n, v

    def do_installed(self):
        for n, v in self.ls():
            yield n , v

    def do_available(self):
        for n, v in self.ls():
            yield n , v

    def do_remove(self, pkg):
        self.run(f'npm uninstall {pkg}')

    def do_cleanup(self):
        pass

    def do_update(self):
        pass

    def do_install(self, pkg, ver=None):
        self.run(f'npm install -g {pkg}')



from setux.core.manage import Manager


class Distro(Manager):
    manager = 'login'

    @property
    def name(self):
        ret, out, err = self.run('id -un')
        return out[0]

    @property
    def id(self):
        ret, out, err = self.run('id -u')
        return int(out[0])



from setux.core.service import Service


class Debian(Service):
    manager = 'SystemD'

    def data(self, svc):
        ret, out, err = self.run(
            f'systemctl -q --no-pager show {svc}',
            report='quiet',
        )
        rows = (i.split('=', 1) for i in out)
        data = dict(i for i in rows if len(i)==2)
        return data

    def do_enabled(self, svc):
        data = self.data(svc)
        state = data.get('UnitFileState')
        return state=='enabled'

    def do_status(self, svc):
        data = self.data(svc)
        active = data['ActiveState']=='active'
        running = data['SubState']=='running'
        return active and running

    def do_start(self, svc):
        self.run(f'systemctl start {svc}')

    def do_stop(self, svc):
        self.run(f'systemctl stop {svc}')

    def do_restart(self, svc):
        self.run(f'systemctl restart {svc}')

    def do_enable(self, svc):
        self.run(f'systemctl enable {svc}')

    def do_disable(self, svc):
        self.run(f'systemctl disable {svc}')


class FreeBSD(Service):
    manager = 'Service'

    def update_conf(self, org, dst):
        cfg = '/etc/rc.conf'
        cont = target.read(cfg)
        if dst in cont: return
        done, updated = False, []
        for line in cont.split('\n'):
            if line == org:
                updated.append(dst)
                done = True
            else:
                updated.append(line)
        if not done:
            updated.append(dst)
        target.write(cfg, '\n'.join(updated))

    def do_status(self, svc):
        ret, out, err = self.run(f'service {svc} status')
        return out[0].startswith(f'{svc} is running')

    def do_start(self, svc):
        self.run(f'service {svc} onestart')

    def do_stop(self, svc):
        self.run(f'service {svc} onestop')

    def do_restart(self, svc):
        self.run(f'service {svc} onerestart')

    def do_enabled(self, svc):
        ret, out, err = self.run(f'service {svc} rcvar')
        return 'YES' in out[0]

    def _do_enable(self, svc):
        self.update_conf(
            org = f'{svc}_enable="NO"',
            dst = f'{svc}_enable="YES"',
        )

    def _do_disable(self, svc):
        self.update_conf(
            org = f'{svc}_enable="YES"',
            dst = f'{svc}_enable="NO"',
        )

    def _do_enable(self, svc):
        self.run(f'sysrc {svc}_enable=YES')

    def do_disable(self, svc):
        self.run(f'sysrc {svc}_enable=NO')




from setux.logger import error
from setux.core.package import CommonPackager


# pylint: disable=no-member


class Distro(CommonPackager):
    manager = 'pip'

    def ls(self):
        ret, out, err = self.run('python3 -m pip list')
        for line in self.out:
            try:
                n, v = line.split()
                v = v.replace('(', '')
                v = v.replace(')', '')
                yield n, v
            except:
                error(line)

    def do_init(self):
        self.run(f'python3 -m pip install -qU pip')

    def do_install(self, pkg, ver=None):
        ret, out, err = self.run(f'python3 -m pip install -qU {pkg}')
        for i in out:
            if ('already satisfied' in i
                or 'Successfully installed' in i
            ):
                break
        else:
            if any(line.strip() for line in out):
                error('\n'.join(out))

    def do_installed(self):
        for n, v in self.ls():
            yield n , v

    def do_available(self):
        for n, v in self.ls():
            yield n , v

    def do_remove(self, pkg):
        self.run(f'python3 -m pip uninstall -y {pkg}')

    def do_cleanup(self):
        pass

    def do_update(self):
        pass




from setux.core.manage import Manager


class Distro(Manager):
    manager = 'os'

    @property
    def kernel(self):
        ret, out, err = self.run('uname -s')
        return out[0]

    @property
    def version(self):
        ret, out, err = self.run('uname -r')
        return out[0].split('-')[0]

    @property
    def arch(self):
        ret, out, err = self.run('uname -m')
        return out[0]



from setux.logger import debug
from setux.core.manage import SpecChecker


class Distro(SpecChecker):
    manager = 'user'

    def get(self):
        user = self.key if self.key else self.spec['uid']
        ret, out, err = self.run(f'grep ^{user}: /etc/passwd')
        for line in out:
            name, x, uid, gid, rem, home, shell = line.split(':')
            if self.key:
                if name != self.key: continue
            else:
                if int(uid) != self.spec['uid']: continue
            return dict(
                name = name,
                uid = int(uid),
                gid = int(gid),
                home = home,
                shell = shell,
            )

    @property
    def group(self):
        return self.distro.group(
            self.key,
            gid = self.spec.get('gid')
        )

    @property
    def home(self):
        spec = self.get()
        home = self.distro.dir(
            spec['home'],
            user = spec['name'],
            group = self.group.name,
        )
        return home

    def chk(self, name, value, spec):
        if name=='home':
            return self.home.check()
        return value == spec

    def cre(self, check=True):
        debug(f'user create {self.key}')
        self.do_cre(user=self.key, check=check)
        self.home.set()

    def mod(self, key, val, check=True):
        debug(f'user {self.key} change {key} -> {val}')
        if key=='gid':
            self.distro.group(self.key, gid=val).set()
        if key=='home':
            self.home.set()
        self.do_mod(user=self.key, key=key, val=val, check=check)

    def rm(self, check=True):
        debug(f'user delete {self.key}')
        self.do_rm(user=self.key, check=check)

    def do_cre(self, *, user, check):
        self.run(f'useradd {user}', check=check)

    def do_mod(self, *, user, key, val, check):
        self.run(f'usermod --{key} {val} {user}', check=check)

    def do_rm(self, *, user, check):
        self.run(f'userdel -r {user}', check=check)


class FreeBSD(Distro):
    opts = dict(
        uid = 'u',
        gid = 'g',
        shell= 's',
        home= 'd',
    )

    def do_cre(self, *, user, check):
        self.run(f'pw useradd -n {user}', check=check)

    def do_mod(self, *, user, key, val, check):
        opt = self.opts[key]
        self.run(f'pw usermod -n {user} -{opt} {val}', check=check)

    def do_rm(self, *, user, check):
        self.run(f'pw userdel -n {user} -r', check=check)




from pybrary.ascii import oct_mod

from setux.logger import debug
from setux.core.manage import SpecChecker


class Distro(SpecChecker):
    manager = 'dir'

    def get(self):
        ret, out, err = self.run(f'ls -ld --color=never {self.key}')
        try:
            mod, ln, usr, grp, size, month, day, time, path = out[0].split()
            typ, mod = mod[0], mod[1:10]
            assert typ=='d', f'DIR {self.key} : {typ} !'
            return dict(
                name = path,
                mode = oct_mod(mod),
                user = usr,
                group = grp,
            )
        except Exception as x:
            debug('dir %s !\n%s : %s', self.key, type(x), x)

    def __str__(self):
        data = self.get()
        if data:
            return f'Dir {self.key} {data["mod"]} {data["usr"]}:{data["grp"]}'
        else:
            return f'Dir {self.key} not found'

    def cre(self):
        debug(f'dir create {self.key}')
        self.run(f'mkdir -pv {self.key}')

    def mod(self, key, val):
        debug(f'dir {self.key} change {key} -> {val}')
        if key=='mode':
            self.run(f'chmod -R {val} {self.key}')
        elif key=='user':
            self.run(f'chown -R {val} {self.key}')
        elif key=='group':
            user = self.spec["user"]
            self.run(f'chown -R {user}:{val} {self.key}')
        else:
            raise KeyError(f' ! {key} !')

    def rm(self):
        debug(f'dir delete {self.key}')
        self.run(f'rm -rf {self.key}')



from pybrary.ascii import rm_ansi_codes

from setux.logger import error, debug
from setux.core.package import SystemPackager


class Debian(SystemPackager):
    manager = 'apt'

    def do_init(self):
        self.do_cleanup()
        self.do_update()

    def do_installed(self):
        ret, out, err = self.run('apt list --installed', report='quiet')
        decolor = rm_ansi_codes
        for line in out:
            name, ver, *_ = line.split()
            name = decolor(name.split('/')[0])
            yield name, ver

    def do_bigs(self):
        ret, out, err = self.target.script('''
            #!/bin/bash
            dpkg-query -Wf '${Installed-Size;9} ${Package}\n' | sort -n | tail -n 22
        ''', report='quiet')
        for line in out:
            yield line

    def do_upgradable(self):
        ret, out, err = self.run('''
            apt list --upgradable
        ''', report='quiet')
        decolor = rm_ansi_codes
        for line in out:
            if not line.strip(): continue
            try:
                line = decolor(line)
                name_ver = line.split()[0]
                name, ver = name_ver.split('/')
                yield name, ver
            except:
                debug(line)

    def do_available(self):
        ret, out, err = self.run('apt list', report='quiet')
        decolor = rm_ansi_codes
        for line in out:
            if '[install' not in line:
                name, ver, *_ = line.split()
                name = decolor(name.split('/')[0])
                yield name, ver

    def do_remove(self, pkg):
        self.run(f'apt-get -y purge {pkg}')

    def do_cleanup(self):
        self.run('apt autoremove -y')

    def do_update(self):
        self.run('apt-get update')

    def do_upgrade(self):
        self.run('apt-get upgrade -y')

    def do_install(self, pkg, ver=None):
        ret, out, err = self.run('apt-get -y install %s%s' % (pkg, '=%s'%ver if ver else ''))
        err = [
            x
            for x in err
            if (
                x.strip()
                and not x.startswith('Extracting templates')
                and x != 'Connection to xyz closed.'
            )
        ]


class FreeBSD(SystemPackager):
    manager = 'pkg'

    def do_init(self):
        self.do_cleanup()

    def do_installed(self):
        cmd = 'pkg query "%n %v"'
        ret, out, err = self.run(f"'{cmd}'", report='quiet')
        for line in out:
            name, ver = line.split()
            yield name, ver

    def do_bigs(self):
        cmd = 'pkg query "%sb %n" | sort -n | tail -n 22'
        ret, out, err = self.run(f"'{cmd}'", report='quiet')
        yield from out

    def do_upgradable(self):
        pass

    def do_available(self):
        cmd = 'pkg search -x ".*"'
        ret, out, err = self.run(f"'{cmd}'", report='quiet')
        for line in out:
            name_ver, *_desc = line.split()
            name, ver = name_ver.rsplit('-', 1)
            yield name, ver

    def do_remove(self, pkg):
        self.run(f'pkg delete -y {pkg}')

    def do_cleanup(self):
        self.run('pkg autoremove -y')
        self.run('pkg clean -y')

    def do_update(self):
        pass

    def do_upgrade(self):
        self.run('pkg upgrade -y')

    def do_install(self, pkg, ver=None):
        ret, out, err = self.run(f'pkg install -y {pkg}')



from setux.logger import debug
from setux.core.manage import ArgsChecker


class Distro(ArgsChecker):
    manager = 'groups'

    def get(self):
        ret, out, err = self.run(f'grep {self.key} /etc/group')
        groups = set()
        for line in out:
            name, x, gid, users = line.split(':')
            if self.key in users.split(','):
                groups.add(name)
        return groups

    def add(self, group, check=True):
        groups = self.get()
        if group not in groups:
            self.distro.group(group).set()
            groups.add(group)
            debug(f'groups add {self.key} to {group}')
            self.do_set(user=self.key, groups=groups, check=check)

    def rm(self, group, check=True):
        groups = self.get()
        if group in groups:
            groups.remove(group)
            debug(f'groups remove {self.key} from {group}')
            self.do_set(user=self.key, groups=groups, check=check)

    def do_set(self, *, user, groups, check):
        self.run(f'usermod --groups {",".join(groups)} {user}', check=check)


class FreeBSD(Distro):
    def do_set(self, *, user, groups, check):
        self.run(f'pw usermod -n {user} -G {",".join(groups)}', check=check)



from pybrary.net import get_ip_adr

from setux.core.manage import Manager


class Distro(Manager):
    manager = 'host'

    @property
    def name(self):
        attr = '_host_name_'
        try:
            val = getattr(self, attr)
        except AttributeError:
            ret, out, err = self.run('hostname')
            val = out[0]
            setattr(self, attr,  val)
        return val

    @name.setter
    def name(self, val):
        attr = '_host_name_'
        delattr(self, attr)
        ret, out, err = self.run(
            'hostname', val.replace('_', '')
        )
        return ret

    @property
    def fqdn(self):
        ret, out, err = self.run('hostname -f')
        return out[0]

    @property
    def addr(self):
        ok, adr = get_ip_adr()
        if ok:
            return adr
        else:
            debug(adr)
            return '!'

